;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; `nokeyb.asm' -- Assembly language source code for `nokeyb.sys'.
;
; This file is part of project `NOKEYB'.
;
; Copyright (c) 2021 Robert Riebisch <rr@bttr-software.de>
;
; Usage of the works is permitted provided that this instrument is retained
; with the works, so that any entity that uses the works is notified of this
; instrument.
;
; DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.
;-----------------------------------------------------------------------------
; Build instructions for NASM version 2.15.05:
;   nasm.exe -o nokeyb.sys nokeyb.asm
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	; limit instruction set
	CPU 8086

	; .SYS file
	ORG 0

deviceDriverHeader:
	DD -1                 ; last driver in file
	;  fedcba9876543210
	DW 1000000000000000b  ; character device driver
.sROffset:
	DW initDriver         ; later patched
	DW interruptRoutine
	;   12345678
	DB "NOKEYB3$"         ; driver name padded to 8 bytes

; device driver status bits
STATUS_ERROR EQU (1 << 15)
STATUS_BUSY  EQU (1 <<  9)
STATUS_DONE  EQU (1 <<  8)

; device driver error code
ERR_UNKNOWN_COMMAND EQU 3

; jump instruction encodings
JMP_SHORT_RELATIVE EQU 0EBh
JMP_FAR_ABSOLUTE   EQU 0EAh

; called with es:bx -> request header
strategyRoutine:
	; just make all calls fail
	mov	word [es:bx + 3], (STATUS_ERROR | STATUS_DONE) + ERR_UNKNOWN_COMMAND

interruptRoutine:
	; return from both routines
	retf

myInt21hHandler:
	push	si
	; intercept function `load and execute' only
	cmp	ax, 4b00h
.jne:
	jne	.notLoadAndExec
	; 'd:\dos\KEYB.COM'
	;  012345678901234
	mov	si, dx
	cmp	word [si], 'd:'
	jne	.notKeyb
	cmp	word [si + 7], 'KE'
	jne	.notKeyb
	cmp	word [si + 9], 'YB'
	jne	.notKeyb
	cmp	word [si + 13], 'OM'
	jne	.notKeyb
	; (ZF=1, CF=0 here)
	pop	si
	; after first interception turn driver off
	mov	byte [cs:.jne], JMP_SHORT_RELATIVE
	; always signal success
	; (CF still 0 from prior `cmp')
	;clc
	; back to caller
	retf	2
.notLoadAndExec:
.notKeyb:
	pop	si
	; let DOS process all other functions
	DB JMP_FAR_ABSOLUTE

oldInt21hAddr:
	; next instruction block takes 4 bytes exactly
	; so recycle!

initDriver:
	; save affected registers
	push	ax
	push	dx
	push	ds
	pushf

.endOfResidentPart:

	; save request header address
	push	es
	push	bx

	; print program name and version
	push	cs
	pop	ds
	mov	dx, msgBanner
	mov	ah, 9
	int	21h

	; hook into DOS services interrupt
	mov	ax, (35h << 8) + 21h
	int	21h
	mov	[oldInt21hAddr], bx
	mov	[oldInt21hAddr + 2], es

	mov	dx, myInt21hHandler
	mov	ah, 25h  ; AL still 21h
	int	21h

	; point to dummy routine to keep system state clear
	mov	word [deviceDriverHeader.sROffset], strategyRoutine

	; restore request header address
	pop	bx
	pop	es

	; signal success to DOS
	mov	word [es:bx + 3], STATUS_DONE

	; set break address to discard anything above from RAM
	mov	word [es:bx + 14], .endOfResidentPart
	mov	[es:bx + 16], cs

	; restore registers
	popf
	pop	ds
	pop	dx
	pop	ax
	retf

msgBanner:
	DB 'NOKEYB R3', 13, 10
	DB 'Copyright (c) 2021 Robert Riebisch', 13, 10
	DB '[Under FAIR LICENSE]', 13, 10
	DB '$'
